/*
Copyright (C) 1994-1995 Apogee Software, Ltd.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/**********************************************************************
   Digitized sound playback via the PC Speaker, using PWM.
   Partially based on code by James R. Rose, along with an explanations
   from differing sources, like these:
   https://bumbershootsoft.wordpress.com/2016/12/10/beyond-beep-boop-mastering-the-pc-speaker/
   http://qzx.com/pc-gpe/speaker.txt

   (c) Copyright 1994 James R. Dose.  All Rights Reserved.
**********************************************************************/

#define STEREO      1
#define SIXTEEN_BIT 2

#define MONO_8BIT    0
#define STEREO_8BIT  ( STEREO )
#define MONO_16BIT   ( SIXTEEN_BIT )
#define STEREO_16BIT ( STEREO | SIXTEEN_BIT )

#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include "dpmi_wrapper.h"
#include "task_man.h"
#include "drivers.h"
#include "driver_pcdigi.h"

#define PCDigi_SampleRate 9000

static int32_t PCDigi_Installed = FALSE;

static char *PCDigi_BufferStart, *PCDigi_BufferEnd;
static char *PCDigi_CurrentBufferEnd;
static int32_t PCDigi_TransferLength = 0;

static char *PCDigi_SoundPtr;
volatile int32_t PCDigi_SoundPlaying;

static task *PCDigi_Timer;

//#define USE_VOLUME_TABLE
#ifdef USE_VOLUME_TABLE
static uint8_t PCDigi_VolumeTable[256];
static int PCDigi_ServiceIntCounter;
#endif

void ( *PCDigi_CallBack )( void );

static int32_t PCDigi_LockMemory(void);
static void PCDigi_UnlockMemory(void);

int32_t PCDigi_ErrorCode = PCDigi_Ok;

int32_t PCDigi_GetError(void)
{
    return PCDigi_ErrorCode;
}

/*---------------------------------------------------------------------
   Function: PCDigi_ErrorString

   Returns a pointer to the error message associated with an error
   number.  A -1 returns a pointer the current error.
---------------------------------------------------------------------*/

const char *PCDigi_ErrorString( int32_t ErrorNumber )
{
    const char *ErrorString;

    switch( ErrorNumber ) {
        case PCDigi_Error :
            ErrorString = PCDigi_ErrorString( PCDigi_ErrorCode );
            break;

        case PCDigi_Ok :
            ErrorString = "PCDigi ok.";
            break;

        case PCDigi_DPMI_Error :
            ErrorString = "DPMI Error in PCDigi.";
            break;

        default :
            ErrorString = "Unknown PCDigi error code.";
            break;
    }

    return ErrorString;
}

/**********************************************************************

   Memory locked functions:

**********************************************************************/

#define PCDigi_LockStart PCDigi_ServiceInterrupt

/*---------------------------------------------------------------------
   Function: PCDigi_ServiceInterrupt

   Handles interrupt generated by sound card at the end of a voice
   transfer.  Calls the user supplied callback function.
---------------------------------------------------------------------*/

static void PCDigi_ServiceInterrupt(task *UNUSED(Task))
{
#ifdef USE_VOLUME_TABLE
    uint8_t temp = inp(0x61);
    outp(0x61, temp | 1);
    outp(0x61, temp & 0xfe);
    if (PCDigi_ServiceIntCounter++ == 0)
       return;
    outp(0x42, PCDigi_VolumeTable[*PCDigi_SoundPtr++]);
    PCDigi_ServiceIntCounter = 0;
#else
    uint8_t scaled = (*PCDigi_SoundPtr++ >= 128) ? 2 : 0;
    outp(0x61, (inp(0x61) & 0xfc) | scaled);
#endif

    if (PCDigi_SoundPtr == PCDigi_CurrentBufferEnd) {
        if (PCDigi_CurrentBufferEnd == PCDigi_BufferEnd)
            PCDigi_SoundPtr = PCDigi_CurrentBufferEnd = PCDigi_BufferStart;
        PCDigi_CurrentBufferEnd += PCDigi_TransferLength;
        if ( PCDigi_CallBack != NULL )
            PCDigi_CallBack();
    }
}

/*---------------------------------------------------------------------
   Function: PCDigi_PCM_StopPlayback

   Ends the transfer of digitized sound to the PC Speaker.
---------------------------------------------------------------------*/

void PCDigi_PCM_StopPlayback(void)
{
    if (PCDigi_SoundPlaying) {
        TS_Terminate(PCDigi_Timer);

        // Turn off speaker
        outp(0x61, inp(0x61) & 0xfc);
#ifdef USE_VOLUME_TABLE
        outp(0x43, 0xb6);
#endif

        PCDigi_SoundPlaying = FALSE;

        PCDigi_SoundPtr = PCDigi_BufferEnd = NULL;
    }
}

/*---------------------------------------------------------------------
   Function: PCDigi_LockEnd

   Used for determining the length of the functions to lock in memory.
---------------------------------------------------------------------*/

static void PCDigi_LockEnd(void) {}

/*---------------------------------------------------------------------
   Function: PCDigi_PCM_BeginPlayback

   Begins multibuffered playback of digitized sound on the PC Speaker.
---------------------------------------------------------------------*/

int32_t PCDigi_PCM_BeginPlayback(char *BufferStart, int32_t BufferSize,
                             int32_t NumDivisions, void ( *CallBackFunc )( void ) )
{
    if (PCDigi_SoundPlaying)
        PCDigi_PCM_StopPlayback();

    PCDigi_CallBack = CallBackFunc;
    PCDigi_BufferStart = PCDigi_SoundPtr = BufferStart;
    PCDigi_CurrentBufferEnd = BufferStart + BufferSize;
    PCDigi_BufferEnd = BufferStart + BufferSize * NumDivisions;
    PCDigi_TransferLength = BufferSize;
    PCDigi_SoundPlaying = TRUE;

#ifdef USE_VOLUME_TABLE
    outp(0x61, inp(0x61) | 3);
    outp(0x43, 0x92);
    PCDigi_Timer = TS_ScheduleTask(PCDigi_ServiceInterrupt, 2*PCDigi_SampleRate, 1, NULL);
#else
    PCDigi_Timer = TS_ScheduleTask(PCDigi_ServiceInterrupt, PCDigi_SampleRate, 1, NULL);
#endif
    TS_Dispatch();

    return PCDigi_Ok;
}


/*---------------------------------------------------------------------
   Function: PCDigi_GetPlaybackRate

   Returns the rate at which the digitized sound will be played in
   hertz.
---------------------------------------------------------------------*/

int32_t PCDigi_GetPlaybackRate(void)
{
    return PCDigi_SampleRate;
}

/*---------------------------------------------------------------------
   Function: PCDigi_SetMixMode

   Sets the sound card to play samples in mono or stereo.
---------------------------------------------------------------------*/

int32_t PCDigi_SetMixMode(int32_t mode)
{
    mode = MONO_8BIT;
    return( mode );
}

/*---------------------------------------------------------------------
   Function: PCDigi_PCM_Init

   Initializes the PC Speaker, prepares the module to play digitized
   sounds.
---------------------------------------------------------------------*/

int32_t PCDigi_PCM_Init(int32_t *mixrate, int32_t *numchannels, void * initdata)
{
    UNREFERENCED_PARAMETER(initdata);

    if (PCDigi_Installed)
        PCDigi_PCM_Shutdown();

    int32_t status = PCDigi_LockMemory();
    if (status != PCDigi_Ok) {
        PCDigi_UnlockMemory();
        return( status );
    }

    status = PCDigi_Ok;
    PCDigi_SoundPlaying = FALSE;
    PCDigi_CallBack = NULL;
    PCDigi_Installed = TRUE;
    PCDigi_ErrorCode = status;
    *mixrate = PCDigi_SampleRate;
    *numchannels = 1;

#ifdef USE_VOLUME_TABLE
    for (int32_t i = 0; i < 256; ++i)
        PCDigi_VolumeTable[i] = (i * 1193182) / (PCDigi_SampleRate * 255 * 2);
#endif

    return status;
}


/*---------------------------------------------------------------------
   Function: PCDigi_PCM_Shutdown

   Ends transfer of sound data to the PC Speaker.
---------------------------------------------------------------------*/

void PCDigi_PCM_Shutdown(void)
{
    PCDigi_PCM_StopPlayback();

    PCDigi_SoundPlaying = FALSE;
    PCDigi_CallBack = NULL;
    PCDigi_UnlockMemory();
    PCDigi_Installed = FALSE;
}


/*---------------------------------------------------------------------
   Function: PCDigi_UnlockMemory

   Unlocks all neccessary data.
---------------------------------------------------------------------*/

static void PCDigi_UnlockMemory(void)
{
    DPMI_UnlockMemoryRegion((void *)PCDigi_LockStart, (void *)PCDigi_LockEnd);
    DPMI_Unlock(PCDigi_Installed);
    DPMI_Unlock(PCDigi_BufferStart);
    DPMI_Unlock(PCDigi_BufferEnd);
    DPMI_Unlock(PCDigi_CurrentBufferEnd);
    DPMI_Unlock(PCDigi_TransferLength);
    DPMI_Unlock(PCDigi_SoundPtr);
    DPMI_Unlock(PCDigi_SoundPlaying);
    DPMI_Unlock(PCDigi_Timer);
#ifdef USE_VOLUME_TABLE
    DPMI_Unlock(PCDigi_VolumeTable);
#endif
    DPMI_Unlock(PCDigi_CallBack);
    DPMI_Unlock(PCDigi_ErrorCode);
}


/*---------------------------------------------------------------------
   Function: PCDigi_LockMemory

   Locks all neccessary data.
---------------------------------------------------------------------*/

static int32_t PCDigi_LockMemory(void)
{
    int32_t status;

    status  = DPMI_LockMemoryRegion((void *)PCDigi_LockStart, (void *)PCDigi_LockEnd);
    status |= DPMI_Lock(PCDigi_Installed);
    status |= DPMI_Lock(PCDigi_BufferStart);
    status |= DPMI_Lock(PCDigi_BufferEnd);
    status |= DPMI_Lock(PCDigi_CurrentBufferEnd);
    status |= DPMI_Lock(PCDigi_TransferLength);
    status |= DPMI_Lock(PCDigi_SoundPtr);
    status |= DPMI_Lock(PCDigi_SoundPlaying);
    status |= DPMI_Lock(PCDigi_Timer);
#ifdef USE_VOLUME_TABLE
    status |= DPMI_Lock(PCDigi_VolumeTable);
#endif
    status |= DPMI_Lock(PCDigi_CallBack);
    status |= DPMI_Lock(PCDigi_ErrorCode);

    if (status != DPMI_Ok) {
        PCDigi_UnlockMemory();
        PCDigi_ErrorCode = PCDigi_DPMI_Error;
        return PCDigi_Error;
    }

    return PCDigi_Ok;
}

void PCDigi_PCM_Lock(void) {}
void PCDigi_PCM_Unlock(void) {}
